/** @author Jakub Chomentowski I1ISI */

#include <boost/circular_buffer.hpp>
#include <iostream>

using namespace std;
using namespace boost;

int main(int argc, char *argv[]){

/**Jest to kod przykadowy majcy na celu pokaza na przykdach zastosowanie
 * Szablonu kontenera circula_buffer */

	/**circular_buffer jest szablonem konteneru podobnym do np std::list z t rnic e zaimplementowano w nim mechanizmy cyklicznago zapisu.
	Przy tworzeniu go podajemy jakiego rodzaju i jak wiele elementw ma w sobie przechowywa.
	Tu uwaga szablon nie rozszerza si, nie zwiksza objtoci zajtej pamici automatycznie gdy prbujemy zapisa zbyt duo elementw.*/

	//Utworzenie bufora cyklicznego na 4 liczby int
	circular_buffer<int> buforCykliczny(4);


	//Do bufora mona zapisywa dane zrwno od przodu
	buforCykliczny.push_front(1);
	buforCykliczny.push_front(0);

	//Jak i od tyu
	buforCykliczny.push_back(3);

	
	//Random acces i odczyt danych zapewnia nam operator "[]"
	cout<<"Przyklad wypisywania z bufora liczb"<<endl;
	cout<<buforCykliczny[0]<<endl;		//wypisze 0
	cout<<buforCykliczny[1]<<endl;		//wypisze 1
	cout<<buforCykliczny[2]<<endl;		//wypisze 2
	cout<<endl;


	//Uwaga: Jeli sprbujemy odczyta co z poza zakresu <0; size()-1> wystpi bd
	//cout<<buforCykliczny[3]<<endl;


	
	//Aby zdj element z bufora naley uy funkcji pop w rnych wariacjach. 
	//Uwaga: Prba zdjcia elementu z bufora pustego zakoczy si rzuceniem wyjtku.

	//pop_front() uswa element z pocztku
	buforCykliczny.pop_front();			//usuwa pierwszy element
	//a pop_back() z konca
	buforCykliczny.pop_back();			//usuwa ostatni element	

	//teraz take index nam si przesunie. W buforze mamy ju tylko jeden element pod indexem 0 i jest to 1
	cout<<"Wypisanie elementow po usunieciu paru sposrod nich"<<endl;
	cout<<buforCykliczny[0]<<endl;		//wypisze 1
	cout<<endl;


	//Jak kady kontener bazujcy na STL circular_buffer zapewnia iteratory umoliwiajce przegldanie swojej zawartoci
	//Naley jednak podkreli e iteratory nie zachowuj si cyklicznie tote betin()-1 i end()+1 s bedn, gdy wskazuj na nieznane miejsce w pamici
	circular_buffer<int>::iterator it = buforCykliczny.begin();
	cout<<"Wypisanie wartosci zapisanej w iteratorze"<<endl;
	cout<<*it<<endl<<endl;				//wypisze 1

	//Naley zwrci uwag i iterator nie przesunie si automatycznie po wstawieniu kolejnego elementu.
	buforCykliczny.push_front(0);
	cout<<"Wypisanie starego iteratora i nowego poczatku bufora"<<endl;
	cout<<buforCykliczny[0]<<endl;		//wypisze 0
	cout<<*it<<endl;					//wypisze 1
	cout<<*(it - 1)<<endl;				//wypisze 0
	cout<<endl;
	
	
	//"cykliczno" bufora ujawnia si wtedy gdy bufor jest zapeniony
	buforCykliczny.push_back(2);
	buforCykliczny.push_back(3);
	cout<<"Pelny bufor"<<endl;
	cout<<buforCykliczny.capacity()<<endl;	//Wypisze ile moe maksymalnie zmieci si w buforze czyli 4
	cout<<buforCykliczny.size()<<endl;		//Wypisze ile elementw jest wewntrz bufora czyli 4
	cout<<endl;

	//Gdy dodam do bufora jeszcze jedna liczb na kocu to nowym pocztkiem bufora bdzie aktualna liczba o indeksie 1
	//a dotychczasowy pocztek zostanie nadpisany now wartoci.
	buforCykliczny.push_back(4);
	cout<<"Przyklad cyklicznosci"<<endl;
	cout<<*(buforCykliczny.end()-1)<<" "<<buforCykliczny[3]<<endl;			//Wypisze 4 4
	cout<<*(buforCykliczny.begin())<<" "<<buforCykliczny[0]<<endl;			//Wypisze 1 1
	cout<<endl;

	//Oczywicie jeli sprobujemy nadpisa element uywajc push_front() analogicznie przesunie nam si koniec bufora
	buforCykliczny.push_front(0);
	cout<<"Przyklad cyklicznosci 2"<<endl;
	cout<<*(buforCykliczny.end()-1)<<" "<<buforCykliczny[3]<<endl;			//Wypisze 3 3
	cout<<*(buforCykliczny.begin())<<" "<<buforCykliczny[0]<<endl;			//Wypisze 0 0
	cout<<endl;

	
	//circular_buffer Ma wbudowane pewne mechanizmy uatwiajce debugowanie. Ma on zapisane wszystkie iteratory do kadego
	//elementu ktry zwrcia. Gdy jaki element jest uswany z bufora take iteratory odpowiadajce mu s uswane z listy.
	//Jeli programista sprbuje uy takiego iteratora zostanie poinformowany przy pomocy assercji o bdzie.
	buforCykliczny.push_back(4);		//Uswa 0 i wstawia na jego miejsce 4
	buforCykliczny.push_back(5);		//Uswa 1 i wstawia na jego miejsce 5
	/*cout<<*it;*/						//Jeli teraz uyje iteratora it ktry wskazuje na element 1 wystpi bd; 



	//UWAGA: Z RACJI CYKLICZNEJ NATURY BUFORA NIE NALEY PRZECHOWYWA W NIM WSKAZA NA AUTOMATYCZNIE ALOKOWANE OBIEKTY, CHYBA E
	//		 UYWANE S SPRYTE WSKANIKI!!!

	//UWAGA2: CIRCULAR_BUFFER NIE JEST BEZPIECZNY PRZY WIELOWATKOWOCI CHYBA E KADY WTEK MA SWOJ WASN INSTANCJ BUFORA LUB
	//		  JELI WTKI TYLKO I WYCZNIE WYKONUJ OPERACJ CZYTANIA. W PRZECIWNYM RAZIE MOG WYSTPI WYCIGI.

	//CIEKAWOSTKA:	Istnieje pewna warjacja tego szablonu mianowicie circular_buffer_space_optimized. Dziaa ona analogicznie do 
	//				zwykego circular_buffer jednak ma t istotn rnic, e nie alokuje od razu bufora o rozmiarze podanym w konstruktorze.
	//				circular_buffer_space_optimized alokuje tylko kawaek potrzebnej pamici, a gdy ta si przepeni alokuje kolejny kawaek.
	//				Dopiero gdy rozmiar bufora osignie zaplanowan wielko obiekt przestaje alokowa kolejne porcje pamici i zaczyna
	//				korzysta z mechanizmw cyklicznego zapisu. Moe to by przydatne gdy tworzymy duy bufor duych elementw, ale istnieje
	//				spora szansa e nie uda nam si go przepeni.
}